(function(angular, ionic, undefined) {

    angular
        .module('headsup')
        .service('$deviceManager', DeviceManager);

    DeviceManager.$inject = ['$cordovaDevice', '$cordovaNetwork', '$cordovaToast', '$http', '$q', '$rootScope', '$utilities', '$configuration', '$categories', '$locations', '$localStorage'];

    function DeviceManager($cordovaDevice, $cordovaNetwork, $cordovaToast, $http, $q, $rootScope, $utilities, $configuration, $categories, $locations, $localStorage) {

        var
            self = this,
            initializationCache = null,
            subscribedLocations = [],
            subscribedCategories = [];

        this.getUuid = getUuid;
        this.getApp = getApp;
        this.getPlatform = getPlatform;
        this.alert = alert;
        this.isMobile = isMobile;
        this.isBrowser = isBrowser;
        this.isIOS = isIOS;
        this.isAndroid = isAndroid;
        this.isOnline = isOnline;
        this.initiate = initiate;
        this.getSubscribedLocations = getSubscribedLocations;
        this.isSubscribedToLocation = isSubscribedToLocation;
        this.subscribeToLocation = subscribeToLocation;
        this.unsubscribeFromLocation = unsubscribeFromLocation;
        this.getSubscribedCategories = getSubscribedCategories;
        this.isSubscribedToCategory = isSubscribedToCategory;
        this.subscribeToCategory = subscribeToCategory;
        this.unsubscribeFromCategory = unsubscribeFromCategory;
        this.isAlertRead = isAlertRead;
        this.setAlertRead = setAlertRead;
        this.setAlertUnread = setAlertUnread;
        this.clearAllReadAlerts = clearAllReadAlerts;
        this.isAlertHidden = isAlertHidden;
        this.hideAlert = hideAlert;
        this.unhideAlert = unhideAlert;
        this.unhideAllAlerts = unhideAllAlerts;
        this.getHiddenAlertsIds = getHiddenAlertsIds;

        /**
         * Ged device unique ID
         *
         * @return {string}
         */
        function getUuid() {

            if (self.isBrowser() && $configuration.get('environment') === 'dev') { // If it is a browser, that means that we are working on development and app is not fired in Cordova. So we will send dummy payload.
                return $configuration.get('uuid');
            } else if (self.isMobile()) {

                if ($configuration.get('uuid')) {
                    $configuration.set('uuid', $cordovaDevice.getUUID());
                };

                return $configuration.get('uuid');

            } else {
                throw 'Unsupported platform: "' + ionic.Platform.platform() + '".';
            };
        };

        /**
         * Get device application
         *
         * @return {string}
         */
        function getApp() {
            return $configuration.get('app');
        };

        /**
         * Get platform name.
         *
         * @return {string}
         * @throws Error
         */
        function getPlatform() {

            if (self.isBrowser() && $configuration.get('environment') === 'dev') { // If it is a browser, that means that we are working on development and app is not fired in Cordova. So we will send dummy payload.
                return $configuration.get('platform');
            };

            if (self.isIOS()) {
                $configuration.set('platform', 'ios');
            } else if (self.isAndroid()) {
                $configuration.set('platform', 'android');
            } else {
                throw 'Unsupported platform: "' + ionic.Platform.platform() + '".';
            };

            return $configuration.get('platform');
        };

        /**
         * Show alert popup message
         *
         * @param {string} message Message to show
         * @param {string} duration Duration of alert to display, 'short'|'long'
         * @param {string} position Position of alert message, 'top'|'center'|'bottom'
         */
        function alert(message, duration, position) {

            if (self.isBrowser()) {
                window.alert(message);
            } else {

                $cordovaToast.show(
                    message,
                    (undefined !== duration) ? duration : 'short',
                    (undefined !== position) ? position : 'center'
                );
            };
        };

        /**
         * Check if device is mobile device
         *
         * @return {boolean}
         */
        function isMobile() {
            return ionic.Platform.isIOS() || ionic.Platform.isAndroid();
        };

        /**
         * Check if device is web browser
         *
         * @return {boolean}
         */
        function isBrowser() {
            return !self.isMobile();
        };

        /**
         * Check if device is IOS device
         *
         * @return {boolean}
         */
        function isIOS() {
            return ionic.Platform.isIOS();
        };

        /**
         * Check if device is android device
         *
         * @return {boolean}
         */
        function isAndroid() {
            return ionic.Platform.isAndroid();
        };

        /**
         * Check if device is online
         *
         * @return {boolean}
         */
        function isOnline() {
            if (self.isBrowser()) {
                return navigator.onLine;
            } else {
                return $cordovaNetwork.isOnline();
            };
        };

        /**
         * Initialize device
         *
         * @return {$q.promise} Promise
         */
        function initiate() {

            var
                deferred = (initializationCache && 'promise' in initializationCache) ? initializationCache.promise : $q.defer();

            if (initializationCache && 'promise' in initializationCache) {

                return initializationCache.promise;

            } else if (initializationCache === null) {

                initializationCache = deferred;

                $http({
                    method: 'POST',
                    url: $utilities.getApiUrl('/device/initiate/v1'),
                    headers: {
                        'Content-Type': 'application/json'
                    },
                    data: {
                        "deviceType": self.getPlatform(),
                        "deviceId": self.getUuid(),
                        "deviceApp": self.getApp()
                    }
                }).then(function(result) {

                    if (result.data.error) {

                        return deferred.reject({
                            message: 'Unable to initiate user.',
                            reason: {
                                message: result.data.message
                            }
                        });

                    } else {

                        subscribedLocations = result.data.subscribedAudiences;
                        subscribedCategories = result.data.subscribedCategories;
                        initializationCache = result.data;

                        $categories.setAll($utilities.processSettingsTree(result.data.categories, 'category'));
                        $locations.setAll($utilities.processSettingsTree(result.data.audiences, 'audience'));

                        return deferred.resolve(result.data);
                    };
                }, function(reason) {

                    return deferred.reject({
                        message: 'Unable to initiate user.',
                        reason: reason
                    });
                });

            } else {
                deferred.resolve(initializationCache);
            };

            return deferred.promise;
        };

        /**
         * Get current subscribed locations IDs
         *
         * @return {int[]}
         */
        function getSubscribedLocations() {
            return subscribedLocations;
        };

        /**
         * Check if device is subscribed to certain location
         *
         * @param {int|string} id Location ID
         * @return {boolean}
         */
        function isSubscribedToLocation(id) {

            var
                result = false;

            angular.forEach(subscribedLocations, function(subscribed) {

                if (subscribed == id) {
                    result = true;
                    return false;
                };
            });

            return result;
        };

        /**
         * Subscribe to certain location
         *
         * @param {int} locationId Location ID to subscribe to
         *
         * @return {$q.promise}
         */
        function subscribeToLocation(locationId) {
            var
                deferred = $q.defer();

            $http({
                method: 'POST',
                url: $utilities.getApiUrl('/audience/subscribe/v1'),
                headers: {
                    'Content-Type': 'application/json'
                },
                data: {
                    "deviceType": self.getPlatform(),
                    "deviceId": self.getUuid(),
                    "deviceApp": self.getApp(),
                    "audienceId": locationId
                }
            }).then(function(result) {

                if (result.data.error) {

                    return deferred.reject({
                        message: 'Unable to subscribe to location.',
                        reason: {
                            message: result.data.message
                        }
                    });

                } else {
                    subscribedLocations = result.data.subscribedAudiences;
                    $rootScope.$broadcast('headsup:locations:changed');

                    return deferred.resolve(result.data);
                };
            }, function(reason) {

                return deferred.reject({
                    message: 'Unable to subscribe to location.',
                    reason: reason
                });

            });

            return deferred.promise;
        };

        /**
         * Unsubscribe from certain location
         *
         * @param {int} locationId Location ID to unsubscribe from
         * @return {$q.promise}
         */
        function unsubscribeFromLocation(locationId) {
            var
                deferred = $q.defer();

            $http({
                method: 'POST',
                url: $utilities.getApiUrl('/audience/unsubscribe/v1'),
                headers: {
                    'Content-Type': 'application/json'
                },
                data: {
                    "deviceType": self.getPlatform(),
                    "deviceId": self.getUuid(),
                    "deviceApp": self.getApp(),
                    "audienceId": locationId
                }
            }).then(function(result) {

                if (result.data.error) {

                    return deferred.reject({
                        message: 'Unable to unsubscribe from location.',
                        reason: {
                            message: result.data.message
                        }
                    });

                } else {
                    subscribedLocations = result.data.subscribedAudiences;
                    $rootScope.$broadcast('headsup:locations:changed');

                    return deferred.resolve(result.data);
                };
            }, function(reason) {

                return deferred.reject({
                    message: 'Unable to unsubscribe from location.',
                    reason: reason
                });
            });

            return deferred.promise;
        };


        /**
         * Get current subscribed categories IDs
         *
         * @return {int[]}
         */
        function getSubscribedCategories() {
            return subscribedCategories;
        };

        /**
         * Check if device is subscribed to certain category
         *
         * @param {int|string} id Category ID
         * @return {boolean}
         */
        function isSubscribedToCategory(id) {

            var
                result = false;

            angular.forEach(subscribedCategories, function(subscribed) {
                if (subscribed == id) {
                    result = true;
                    return false;
                };
            });

            return result;
        };

        /**
         * Subscribe to certain category
         *
         * @param {int} categoryId Category ID to subscribe to
         *
         * @return {$q.promise}
         */
        function subscribeToCategory(categoryId) {
            var
                deferred = $q.defer();

            $http({
                method: 'POST',
                url: $utilities.getApiUrl('/category/subscribe/v1'),
                headers: {
                    'Content-Type': 'application/json'
                },
                data: {
                    "deviceType": self.getPlatform(),
                    "deviceId": self.getUuid(),
                    "deviceApp": self.getApp(),
                    "categoryId": categoryId
                }
            }).then(function(result) {

                if (result.data.error) {

                    return deferred.reject({
                        message: 'Unable to subscribe to category.',
                        reason: {
                            message: result.data.message
                        }
                    });
                } else {

                    subscribedCategories = result.data.subscribedCategories;
                    $rootScope.$broadcast('headsup:categories:changed');

                    return deferred.resolve(result.data);
                };
            }, function(reason) {

                return deferred.reject({
                    message: 'Unable to subscribe to category.',
                    reason: reason
                });
            });

            return deferred.promise;
        };

        /**
         * Unsubscribe from certain category
         *
         * @param {int} categoryId Category ID to unsubscribe from
         * @return {$q.promise}
         */
        function unsubscribeFromCategory(categoryId) {
            var
                deferred = $q.defer();

            $http({
                method: 'POST',
                url: $utilities.getApiUrl('/category/unsubscribe/v1'),
                headers: {
                    'Content-Type': 'application/json'
                },
                data: {
                    "deviceType": self.getPlatform(),
                    "deviceId": self.getUuid(),
                    "deviceApp": self.getApp(),
                    "categoryId": categoryId
                }
            }).then(function(result) {

                if (result.data.error) {

                    return deferred.reject({
                        message: 'Unable to remove category.',
                        reason: {
                            message: result.data.message
                        }
                    });

                } else {

                    subscribedCategories = result.data.subscribedCategories;
                    $rootScope.$broadcast('headsup:categories:changed');

                    return deferred.resolve(result.data);
                };
            }, function(reason) {

                return deferred.reject({
                    message: 'Unable to remove category.',
                    reason: reason
                });
            });

            return deferred.promise;
        };

        /**
         * Check if certain alert is read.
         *
         * @param {int} alertId
         * @return {boolean}
         */
        function isAlertRead(alertId) {

            var
                read = $localStorage.get('alert.read', []);

            return read.indexOf(parseInt(alertId)) !== -1;
        };

        /**
         * Set that alert is read on this device.
         *
         * @param {int} alertId
         * @return {DeviceManager}
         */
        function setAlertRead(alertId) {
            var
                read = $localStorage.get('alert.read', []);

            read.push(parseInt(alertId));

            $localStorage.save('alert.read', read);

            return self;
        };

        /**
         * Set that alert is unread on this device.
         *
         * @param {int} alertId
         * @return {DeviceManager}
         */
        function setAlertUnread(alertId) {
            alertId = parseInt(alertId);

            var
                read = $localStorage.get('alert.read', []),
                index = read.indexOf(alertId);

            if (index !== -1) {
                read.splice(index, 1);
                $localStorage.save('alert.read', read);
            };

            return self;
        };

        /**
         * Clear all read flags.
         *
         * @return {DeviceManager}
         */
        function clearAllReadAlerts() {
            $localStorage.save('alert.read', []);
            return self;
        };

        /**
         * Check if alert is hidden.
         *
         * @param {int} alertId
         * @return {boolean}
         */
        function isAlertHidden(alertId) {
            var
                read = $localStorage.get('alert.hidden', []);

            return read.indexOf(parseInt(alertId)) !== -1;
        };

        /**
         * Hide alert.
         *
         * @param {int} alertId
         * @return {DeviceManager}
         */
        function hideAlert(alertId) {
            var
                read = $localStorage.get('alert.hidden', []);

            read.push(parseInt(alertId));

            $localStorage.save('alert.hidden', read);

            return self;
        };

        /**
         * Unhide alert.
         *
         * @param {int} alertId
         * @return {DeviceManager}
         */
        function unhideAlert(alertId) {

            alertId = parseInt(alertId);

            var
                read = $localStorage.get('alert.hidden', []),
                index = read.indexOf(alertId);

            if (index !== -1) {
                read.splice(index, 1);
                $localStorage.save('alert.hidden', read);
            };

            return self;
        };

        /**
         * Unhide all alerts.
         *
         * @return {DeviceManager}
         */
        function unhideAllAlerts() {
            $localStorage.save('alert.hidden', []);
            return self;
        };

        /**
         * Get ids of hidden alerts.
         *
         * @return {Array}
         */
        function getHiddenAlertsIds() {
            return $localStorage.get('alert.hidden', []);
        };
    };


})(angular, ionic);