(function(angular, ionic, window, undefined) {

    'use strict';

    function extractFilenameFromUrl(url) {

        //this removes the anchor at the end, if there is one
        url = url.substring(0, (url.indexOf("#") == -1) ? url.length : url.indexOf("#"));

        //this removes the query after the file name, if there is one
        url = url.substring(0, (url.indexOf("?") == -1) ? url.length : url.indexOf("?"));

        //this removes everything before the last slash in the path
        url = url.substring(url.lastIndexOf("/") + 1, url.length);

        return url;
    };

    function generateStringHash(str) {

        var
            hash = 0;

        if (str.length == 0) return hash;

        for (var i = 0; i < str.length; i++) {
            var
                char = str.charCodeAt(i);

            hash = ((hash<<5)-hash)+char;
            hash = hash & hash; // Convert to 32bit integer
        };

        return hash;
    };

    function generateFilenameFromUrl(url) {
        return generateStringHash(url) + '.' + extractFilenameFromUrl(url);
    };

    /**
     * Storage abstraction
     *
     * @constructor
     */
    function FileStorage($q, $http, $cordovaFile, $cordovaFileTransfer) {

        var
            self = this,
            settings = {
                storage: null
            };

        if (ionic.Platform.isAndroid()) {
            settings.storage = cordova.file.externalDataDirectory;
        } else if (ionic.Platform.isIOS()) {
            settings.storage = cordova.file.cacheDirectory;
        };

        /**
         * Download file to filesystem.
         *
         * @param {string} url URL of the file
         * @return {$q.promise}
         */
        this.download = function(url) {

            var
                deferred = $q.defer(),
                path = self.path(url);

            $cordovaFileTransfer.
                download(url, path, {}, true)
                .then(function() {
                    return deferred.resolve(path);
                })
                .catch(function() {
                    return deferred.reject();
                });

            return deferred.promise;
        };

        /**
         * Check if file exists on give URL
         *
         * @param {string} url URL of the file
         * @return {$q.promise}
         */
        this.has = function(url) {

            var
                deferred = $q.defer(),
                filename = generateFilenameFromUrl(url);

            $cordovaFile
                .checkFile(settings.storage, filename)
                .then(function (success) {
                    return deferred.resolve(true);
                })
                .catch(function (error) {
                    return (
                        (error && error.code == FileError.NOT_FOUND_ERR)
                            ?
                            deferred.resolve(false)
                            :
                            deferred.reject(error)
                    );
                });

            return deferred.promise;
        };

        /**
         * Get local file path based on source url.
         *
         * @param {string} url URL of the file
         * @return {string}
         */
        this.path = function(url) {
            return settings.storage + generateFilenameFromUrl(url);
        };

        /**
         * Get mime of file on given url
         *
         * @param {string} url URL of the file
         * @return {$q.promise}
         */
        this.mime = function(url) {
            var
                deferred = $q.defer();

            window.resolveLocalFileSystemURI(generateFilenameFromUrl(url), function(fileEntry) {
                fileEntry.file(function(entry) {
                    deferred.resolve(entry.type);
                }, function() {
                    deferred.reject();
                });

            }, function() {
                deferred.reject();
            });

            return deferred.promise;
        };
    };

    angular.module('remoteFileOpener', ['ngCordova'])

        .service('$remoteFileOpener', ['$http', '$q', '$cordovaFile', '$cordovaFileTransfer', '$cordovaFileOpener2', function($http, $q, $cordovaFile, $cordovaFileTransfer, $cordovaFileOpener2) {

            var
                self = this,
                storage = new FileStorage($q, $http, $cordovaFile, $cordovaFileTransfer);

            this.has = function(url) {
                return storage.has(url);
            };

            this.download = function(url) {
                return storage.download(url);
            };

            this.open = function(url, mime) {

                var
                    deferred = $q.defer();

                storage
                    .has(url)
                    .then(function(has) {

                        if (has) {
                            return $q.resolve(storage.path(url))
                        } else {
                            return storage.download(url);
                        };
                    })
                    .then(function(path) {

                        var
                            deferred = $q.defer();

                        if (undefined === mime) {

                            storage
                                .mime(url)
                                .then(function(mime) {
                                    return deferred.resolve({
                                        path: path,
                                        mime: mime
                                    });
                                })
                                .catch(function() {
                                    return deferred.resolve({
                                        path: path,
                                        mime: null
                                    });
                                });
                        } else {
                            deferred.resolve({
                                path: path,
                                mime: mime
                            });
                        };

                        return deferred.promise;
                    })
                    .then(function(data) {

                        var
                            deferred = $q.defer();

                        $cordovaFileOpener2
                            .open(data.path, data.mime)
                            .then(function() {
                                return deferred.resolve();
                            }, function() {
                                return deferred.reject();
                            });

                        return deferred.promise;
                    })
                    .catch(function() {
                        return deferred.reject();
                    });

                return deferred.promise;
            };
        }]);

})(angular, ionic, window);
