(function(angular, ionic, undefined) {

    angular
        .module('headsup.alerts')
        .service('$alerts', AlertsService);

    AlertsService.$inject = ['$http', '$q', '$utilities', '$deviceManager', '$categories', '$locations', '$sessionStorage'];

    function AlertsService($http, $q, $utilities, $deviceManager, $categories, $locations, $sessionStorage) {

        this.list = list;
        this.latest = latest;
        this.listMy = listMy;
        this.get = get;
        this.getMy = getMy;
        this.map = map;
        this.publish = publish;
        this.unpublish = unpublish;
        this.save = save;
        this.new = createNew;
        this.loadMetadata = loadMetadata;

        function list(options) {

            var
                deferred = $q.defer();

            if (undefined === options) {
                options = {};
            };

            options = angular.extend({
                offset: 0,
                limit: 30,
                'categories[]': $deviceManager.getSubscribedCategories(),
                'audiences[]': $deviceManager.getSubscribedLocations(),
                'exclude[]': $deviceManager.getHiddenAlertsIds()
            }, options);

            $http({
                method: 'GET',
                url: $utilities.getApiUrl('/alerts/v1'),
                params: options
            }).then(function(result) {

                processAlertDates(result.data.data);

                return deferred.resolve(result.data);

            }, function(reason) {
                return deferred.reject('Unable to get list of alerts. Reason: ' + reason);
            });

            return deferred.promise;
        };

        function latest(latestDate, options) {

            var
                deferred = $q.defer();

            if (undefined === options) {
                options = {};
            };

            options = angular.extend({
                latestDate: latestDate,
                offset: 0,
                limit: 30,
                'categories[]': $deviceManager.getSubscribedCategories(),
                'audiences[]': $deviceManager.getSubscribedLocations(),
                'exclude[]': $deviceManager.getHiddenAlertsIds()
            }, options);

            $utilities.objectPropertiesDateToString(options, 'latestDate');

            $http({
                method: 'GET',
                url: $utilities.getApiUrl('/alerts/latest/v1'),
                params: options
            }).then(function(result) {

                processAlertDates(result.data.data);

                return deferred.resolve(result.data);

            }, function(reason) {
                return deferred.reject('Unable to get latest list of alerts. Reason: ' + reason);
            });

            return deferred.promise;

        };

        function listMy(options) {

            var
                deferred = $q.defer();

            if (undefined === options) {
                options = {};
            };

            options = angular.extend({
                offset: 0,
                limit: 30,
                'categories[]': $categories.ids(),
                'audiences[]': $locations.ids()
            }, options);

            $http({
                method: 'GET',
                url: $utilities.getApiUrl('/alerts/my-alerts/v1'),
                params: options
            }).then(function(result) {

                processAlertDates(result.data.data);

                return deferred.resolve(result.data);

            }, function(reason) {
                return deferred.reject('Unable to get list of alerts. Reason: ' + reason);
            });

            return deferred.promise;
        };

        function get(id) {

            var
                deferred = $q.defer();

            $http({
                method: 'GET',
                url: $utilities.getApiUrl('/alert/v1'),
                params: {
                    id: id
                }
            }).then(function(result) {

                processAlertDates([result.data]);

                return deferred.resolve(result.data);

            }, function(reason) {
                return deferred.reject('Unable to get requested alert. Reason: ' + reason);
            });

            return deferred.promise;
        };

        function getMy(id) {

            var
                deferred = $q.defer();

            $http({
                method: 'GET',
                url: $utilities.getApiUrl('/my-alert/v1'),
                params: {
                    id: id
                }
            }).then(function(result) {

                processAlertDates([result.data]);

                return deferred.resolve(result.data);

            }, function(reason) {
                return deferred.reject('Unable to get requested alert. Reason: ' + reason);
            });

            return deferred.promise;
        };

        function map() {

            var
                deferred = $q.defer();

            $http({
                method: 'GET',
                url: $utilities.getApiUrl('/alert/map/v1'),
                params: {
                    'categories[]': $deviceManager.getSubscribedCategories(),
                    'audiences[]': $deviceManager.getSubscribedLocations(),
                    'exclude[]': $deviceManager.getHiddenAlertsIds()
                }
            }).then(function(result) {

                var
                    geomarkers = [];

                processAlertDates(result.data);

                angular.forEach(result.data, function(alert) {

                    var
                        marker = $categories.marker(alert.category);

                    geomarkers.push(angular.extend($utilities.getCenterOfGeomarkersCollection(alert.geomarkers), {
                        alert: alert,
                        icon: (marker) ? new google.maps.MarkerImage(marker, null, null, null, new google.maps.Size(70, 69)) : null,
                        type: 'marker'
                    }));
                });


                return deferred.resolve(geomarkers);

            }, function(reason) {
                return deferred.reject('Unable to get list of geomarkers for alerts. Reason: ' + reason);
            });

            return deferred.promise;
        };

        function publish(id) {

            var
                deferred = $q.defer();

            $http({
                method: 'POST',
                url: $utilities.getApiUrl('/alert/publish/v1'),
                params: {
                    'id': id
                }
            }).then(function(result) {
                return deferred.resolve();
            }, function(reason) {
                return deferred.reject('Unable to publish alert. Reason: ' + reason);
            });

            return deferred.promise;
        };

        function unpublish(id) {

            var
                deferred = $q.defer();

            $http({
                method: 'POST',
                url: $utilities.getApiUrl('/alert/unpublish/v1'),
                params: {
                    'id': id
                }
            }).then(function(result) {
                return deferred.resolve();
            }, function(reason) {
                return deferred.reject('Unable to publish alert. Reason: ' + reason);
            });

            return deferred.promise;
        };

        function save(alert) {

            var
                deferred = $q.defer(),
                newAlert = createNew();

            for (var prop in newAlert) {

                if (newAlert.hasOwnProperty(prop)) {
                    newAlert[prop] = alert[prop];
                };
            };

            prepareAlertDates([newAlert]);

            $http({
                method: 'POST',
                url: $utilities.getApiUrl('/alert/save/v1'),
                headers: {
                    'Content-Type': 'application/json'
                },
                data: newAlert
            }).then(function(result) {

                if (result.data.error) {
                    return deferred.reject('Unable to save alert. Reason: ' + result.data.message);
                };

                return deferred.resolve(result.data.id);
            }, function(reason) {
                return deferred.reject('Unable to save alert. Reason: ' + reason);
            });

            return deferred.promise;
        };

        function createNew() {

            return {
                id: null,
                title: null,
                category: null,
                audience: ($locations.ids().length === 1) ? [$locations.ids()[0]] : [],
                created: new Date(),
                details: null,
                geomarkers: [],
                images: [],
                files: [],
                web: null,
                type: null,
                status: null,
                urgency: null,
                severity: null,
                certainty: null,
                scope: null,
                restriction: null,
                scopeAddress: [],
                responseType: [],
                eventOnset: null,
                eventExpires: null,
                note: null,
                source: null,
                instruction: [],
                reference: [],
                oasisSender: null,
                oasisIdentifier: null
            };
        };

        function loadMetadata() {

            var
                deferred = $q.defer();

            if ($sessionStorage.has('alert.metadata')) {
                deferred.resolve($sessionStorage.get('alert.metadata'));
            } else {

                $http({
                    method: 'GET',
                    url: $utilities.getApiUrl('/alert/meta/v1'),
                    params: {
                        'deviceApp': $deviceManager.getApp()
                    }
                }).then(function(result) {
                    $sessionStorage.save('alert.metadata', result.data);
                    return deferred.resolve($sessionStorage.get('alert.metadata'));
                }, function(reason) {
                    return deferred.reject('Unable to get alert meta. Reason: ' + reason);
                });
            };

            return deferred.promise;
        };


        function processAlertDates(alerts) {

            angular.forEach(alerts, function(alert) {
                $utilities.objectPropertiesStringToDate(alert, ['created', 'eventOnset', 'eventExpires']);
            });
        };

        function prepareAlertDates(alerts) {
            angular.forEach(alerts, function(alert) {
                $utilities.objectPropertiesDateToString(alert, ['created', 'eventOnset', 'eventExpires']);
            });
        };
    };


})(angular, ionic);

