(function(window, undefined) {

    /**
     * Default settings for geomarkable map.
     *
     * @type {{zoom: number, lat: number, lng: number, mapTypeControl: boolean, editable: boolean}}
     */
    GeomarkableMap.defaults = {
        zoom: 17,
        lat: 0,
        lng: 0,
        mapTypeControl: false,
        editable: false
    };

    /**
     * Drawing modes for geomarkable map.
     *
     * @type {{MOVE: null, MARKER: string, CIRCLE: string, POLYGON: string, POLYLINE: string, RECTANGLE: string}}
     */
    GeomarkableMap.drawingMode = {
        MOVE: null,
        MARKER: 'marker',
        CIRCLE: 'circle',
        POLYGON: 'polygon',
        POLYLINE: 'polyline',
        RECTANGLE: 'rectangle'
    };

    /**
     * A constructor.
     *
     * @param element
     * @param settings
     *
     * @constructor
     */
    function GeomarkableMap(element, settings) {

        /**
         * @type {GeomarkableMap}
         */
        var
            /**
             * Reference to current object instance.
             *
             * @type {GeomarkableMap}
             * @private
             */
            _self = this,

            /**
             * Settings.
             *
             * @type {{zoom: null, lat: null, lng: null, mapTypeControl: null, editable: null}}
             * @private
             */
            _settings = {
                zoom: null,
                lat: null,
                lng: null,
                mapTypeControl: null,
                editable: null
            },
            /**
             * Collection of event listeners
             *
             * @type {{geomarkerAdded: Array, geomarkerSelected: Array, geomarkerDeselected: Array, geomarkerRemoved: Array, geomarkersCleared: Array, drawingModeChanged: Array, geomarkerClicked: Array}}
             * @private
             */
            _eventListeners = {
                'geomarkerAdded': [],
                'geomarkerSelected': [],
                'geomarkerDeselected': [],
                'geomarkerRemoved': [],
                'geomarkersCleared': [],
                'drawingModeChanged': [],
                'geomarkerClicked': []
            },
            /**
             * Google map.
             *
             * {google.maps.Map}
             * @private
             */
            _map = null,
            /**
             * Drawing manager.
             *
             * @type {null|google.maps.drawing.DrawingManager}
             * @private
             */
            _drawingManager = null,
            /**
             * Collection of google maps shapes.
             *
             * @type {Array}
             * @private
             */
            _shapes = [],
            /**
             * Currently selected shape.
             *
             * @type {object|null}
             * @private
             */
            _currentSelection = null;

        element = element.length ? element[0] : element;

        this.addEventListener = addEventListener;
        this.trigger = trigger;
        this.clearEventListeners = clearEventListeners;
        this.setGeomarkers = setGeomarkers;
        this.addMarker = addMarker;
        this.addCircle = addCircle;
        this.addPolygon = addPolygon;
        this.addRectangle = addRectangle;
        this.clear = clear;
        this.enterDrawingMode = enterDrawingMode;
        this.select = select;
        this.deselect = deselect;
        this.hasSelection = hasSelection;
        this.deleteSelected = deleteSelected;
        this.setZoom = setZoom;
        this.setCenter = setCenter;
        this.setViewport = setViewport;
        this.fitCurrentShapes = fitCurrentShapes;
        this.getConfiguration = getConfiguration;
        this.setConfiguration = setConfiguration;
        this.resetConfiguration = resetConfiguration;

        /**
         * Initialize Geomarkable map.
         */
        init();

        /**
         * Add event listener.
         *
         * @param {string} name Event name.
         * @param {function} handler Event handler.
         * @return {GeomarkableMap}
         */
        function addEventListener(name, handler) {

            if (!_eventListeners[name]) {
                throw 'Unknown event name: ' + name;
            };

            _eventListeners[name].push(handler);

            return _self;
        };

        /**
         * Trigger event handler.
         *
         * @param {string} name Event name.
         * @param {Array} args Event arguments.
         * @return {GeomarkableMap}
         */
        function trigger(name, args) {

            for (var i = 0; i < _eventListeners[name].length; i++) {
                _eventListeners[name][i].apply(_self, args);
            };

            return _self;
        };

        /**
         * Clear all event listners.
         *
         * @return {GeomarkableMap}
         */
        function clearEventListeners() {
            _eventListeners = {
                'geomarkerAdded': [],
                'geomarkerSelected': [],
                'geomarkerDeselected': [],
                'geomarkerRemoved': [],
                'geomarkersCleared': [],
                'drawingModeChanged': [],
                'geomarkerClicked': []
            }

            return _self;
        };

        /**
         * Set and render geomarkers on map. Clears current geomarkers.
         *
         * @param {Array} geomarkers Geomarkers to render
         * @return {GeomarkableMap}
         */
        function setGeomarkers(geomarkers) {

            _self.clear();

            if (geomarkers) {

                for (var i = 0; i < geomarkers.length; i++) {

                    var
                        methodName = 'add' + geomarkers[i].type.charAt(0).toUpperCase() +  geomarkers[i].type.toLowerCase().slice(1);

                    if (!_self[methodName]) {
                        throw  'Unknown geomarker type: ' + geomarkers[i].type;
                    };

                    _self[methodName](geomarkers[i]);
                };
            };

            return _self;
        };

        /**
         * Add marker to map.
         *
         * @param {object} geomarker Geomarker
         * @return {GeomarkableMap}
         */
        function addMarker(geomarker) {

            if (undefined !== geomarker.type && null !== geomarker.type && google.maps.drawing.OverlayType.MARKER !== geomarker.type) {
                throw 'You can not add geomarker "' + geomarker.type + '" as marker geomarker type.';
            };

            var
                shape = new google.maps.Marker({
                    position: geomarker,
                    map: _map,
                    draggable: _settings.editable,
                    icon: (geomarker.icon) ? geomarker.icon : null
                });

            shape.geomarker = geomarker;
            shape.type = geomarker.type = google.maps.drawing.OverlayType.MARKER;

            google.maps.event.addListener(shape, 'click', function() {
                _self.select(shape);
            });

            _shapes.push(shape);

            _self.trigger('geomarkerAdded', [shape.geomarker]);

            shape.addListener('dragend', function(event) {
                shape.geomarker.lat = event.latLng.lat();
                shape.geomarker.lng = event.latLng.lng();
            });

            return _self;
        };

        /**
         * Add circle to map.
         *
         * @param {object} geomarker Geomarker
         * @return {GeomarkableMap}
         */
        function addCircle(geomarker) {

            if (undefined !== geomarker.type && null !== geomarker.type && google.maps.drawing.OverlayType.CIRCLE !== geomarker.type) {
                throw 'You can not add geomarker "' + geomarker.type + '" as circle geomarker type.';
            };

            var
                shape = new google.maps.Circle({
                    center: geomarker.center,
                    radius: geomarker.radius,
                    map: _map,
                    editable: _settings.editable
                });

            shape.geomarker = geomarker;
            shape.type = geomarker.type = google.maps.drawing.OverlayType.CIRCLE;

            google.maps.event.addListener(shape, 'click', function() {
                _self.select(shape);
            });

            _shapes.push(shape);

            _self.trigger('geomarkerAdded', [shape.geomarker]);

            google.maps.event.addListener(shape, 'radius_changed', function() {
                this.geomarker.radius = this.getRadius();
            });

            google.maps.event.addListener(shape, 'center_changed', function() {
                this.geomarker.center = {
                    lat: this.center.lat(),
                    lng: this.center.lng()
                };
            });

            return _self;
        };

        /**
         * Add polygon to map.
         *
         * @param {object} geomarker Geomarker
         * @return {GeomarkableMap}
         */
        function addPolygon(geomarker) {

            if (undefined !== geomarker.type && null !== geomarker.type && google.maps.drawing.OverlayType.POLYGON !== geomarker.type) {
                throw 'You can not add geomarker "' + geomarker.type + '" as polygon geomarker type.';
            };

            var
                shape = new google.maps.Polygon({
                    paths: geomarker.points,
                    map: _map,
                    editable: _settings.editable
                });

            shape.geomarker = geomarker;
            shape.type = geomarker.type = google.maps.drawing.OverlayType.POLYGON;

            google.maps.event.addListener(shape, 'click', function() {
                _self.select(shape);
            });

            _shapes.push(shape);

            _self.trigger('geomarkerAdded', [shape.geomarker]);

            google.maps.event.addListener(shape.getPath(), 'set_at', updatePaths);
            google.maps.event.addListener(shape.getPath(), 'remove_at', updatePaths);
            google.maps.event.addListener(shape.getPath(), 'insert_at', updatePaths);

            function updatePaths() {

                shape.geomarker.points = [];

                var
                    paths = shape.getPaths().getArray()[0].getArray();

                for (var i = 0; i < paths.length; i++) {
                    shape.geomarker.points.push({
                        lat: paths[i].lat(),
                        lng: paths[i].lng()
                    });
                };
            };

            return _self;
        };

        /**
         * Add rectangle to map.
         *
         * @param {object} geomarker Geomarker
         * @return {GeomarkableMap}
         */
        function addRectangle(geomarker) {

            if (undefined !== geomarker.type && null !== geomarker.type && google.maps.drawing.OverlayType.RECTANGLE !== geomarker.type) {
                throw 'You can not add geomarker "' + geomarker.type + '" as rectangle geomarker type.';
            };

            var
                shape = new google.maps.Rectangle({
                    bounds: {
                        north: geomarker.northEast.lat,
                        south: geomarker.southWest.lat,
                        east: geomarker.northEast.lng,
                        west: geomarker.southWest.lng
                    },
                    map: _map,
                    editable: _settings.editable
                });

            shape.geomarker = geomarker;
            shape.type = geomarker.type = google.maps.drawing.OverlayType.RECTANGLE;

            google.maps.event.addListener(shape, 'click', function() {
                _self.select(shape);
            });

            _shapes.push(shape);

            _self.trigger('geomarkerAdded', [shape.geomarker]);

            google.maps.event.addListener(shape, 'bounds_changed', function() {
                this.geomarker.northEast = {
                    lat: this.getBounds().getNorthEast().lat(),
                    lng: this.getBounds().getNorthEast().lng()
                };
                this.geomarker.southWest = {
                    lat: this.getBounds().getSouthWest().lat(),
                    lng: this.getBounds().getSouthWest().lng()
                };
            });

            return _self;
        };

        /**
         * Clear all geomarkers.
         *
         * @return {GeomarkableMap}
         */
        function clear() {

            for (var i = 0; i < _shapes.length; i++) {
                _shapes[i].setMap(null);
            };

            _shapes = [];
            _self.trigger('geomarkersCleared');

            return _self;
        };

        /**
         * Enter drawing mode.
         *
         * @param {string} mode
         * @return {GeomarkableMap}
         */
        function enterDrawingMode(mode) {

            if (_drawingManager && mode !== _drawingManager.getDrawingMode()) {

                _drawingManager.setDrawingMode(mode);

                if (mode === GeomarkableMap.drawingMode.MOVE) {
                    _self.deselect();
                };

                _self.trigger('drawingModeChanged', [mode]);
            };

            return _self;
        };

        /**
         * Select drawn shape on map.
         *
         * @param shape
         * @return {GeomarkableMap}
         */
        function select(shape) {

            if (_drawingManager && GeomarkableMap.drawingMode.MOVE !== _drawingManager.getDrawingMode()) {
                _self.enterDrawingMode(GeomarkableMap.drawingMode.MOVE);
            };

            if (shape !== _currentSelection) {
                _currentSelection = shape;
                _self.trigger('geomarkerSelected', [shape.geomarker]);
            };

            _self.trigger('geomarkerClicked', [shape.geomarker]);

            return _self;
        };

        /**
         * Deselect currently selected shape on map.
         *
         * @return {GeomarkableMap}
         */
        function deselect() {

            if (_drawingManager && GeomarkableMap.drawingMode.MOVE !== _drawingManager.getDrawingMode()) {
                _self.enterDrawingMode(GeomarkableMap.drawingMode.MOVE);
            };

            if (null !== _currentSelection) {
                var
                    selected = _currentSelection;

                _currentSelection = null;

                _self.trigger('geomarkerDeselected', [selected.geomarker]);
            };

            return _self;
        };

        /**
         * Check if there is a selected shape on map.
         *
         * @return {boolean}
         */
        function hasSelection() {
            return null !== _currentSelection;
        };

        /**
         * Delete currently selected shape on map.
         *
         * @return {GeomarkableMap}
         */
        function deleteSelected() {

            if (_self.hasSelection()) {

                var
                    selected = _currentSelection;

                selected.setMap(null);
                _self.deselect();

                var
                    index = _shapes.indexOf(selected);

                if (index > -1) {
                    _shapes.splice(index, 1);
                    _self.trigger('geomarkerRemoved', [selected.geomarker]);
                };
            };

            return _self;
        };

        /**
         * Set map zoom.
         *
         * @param {int} zoom Zoom value.
         * @return {GeomarkableMap}
         */
        function setZoom(zoom) {
            _map.setZoom(zoom);
            return _self;
        };

        /**
         * Set map center.
         *
         * @param {number|google.maps.LatLng} lat
         * @param {number|undefined} lng
         * @return {GeomarkableMap}
         */
        function setCenter(lat, lng) {

            if (undefined === lng) {
                _map.setCenter(lat);
            } else {
                _map.setCenter(new google.maps.LatLng({
                    lat: lat,
                    lng: lng
                }));
            };

            return _self;
        };

        /**
         * Set map viewport.
         *
         * @param {google.maps.LatLngBounds} bounds
         * @return {GeomarkableMap}
         */
        function setViewport(bounds) {
            _map.fitBounds(bounds);
            return _self;
        };

        /**
         * Will display map to fit all geomarkers.
         *
         * If there are no geomarkers, it will center and zoom map as per current configuration.
         *
         * If there are geomarkers, it will fit them and zoom to either fit them all, or zoom to minimum
         * configured value.
         *
         * @return {GeomarkableMap}
         */
        function fitCurrentShapes(minimumZoom) {

            if (undefined === minimumZoom) {
                minimumZoom = _settings.zoom;
            };

            google.maps.event.trigger(_map, 'resize');

            if (_shapes.length === 0) {

                setTimeout(function() {
                    _self
                        .setCenter(_settings.lat, _settings.lng)
                        .setZoom(minimumZoom);
                }, 300);

                return _self;
            };

            var
                bounds = new google.maps.LatLngBounds();

            for (var i = 0; i < _shapes.length; i++) {

                switch (_shapes[i].type) {
                    case google.maps.drawing.OverlayType.MARKER:
                        bounds.extend(new google.maps.LatLng(_shapes[i].geomarker));
                        break;
                    case google.maps.drawing.OverlayType.CIRCLE:
                        bounds.extend(_shapes[i].getBounds().getNorthEast());
                        bounds.extend(_shapes[i].getBounds().getSouthWest());
                        break;
                    case google.maps.drawing.OverlayType.POLYGON:
                        for (var j = 0; j < _shapes[i].geomarker.points.length; j++) {
                            bounds.extend(new google.maps.LatLng(_shapes[i].geomarker.points[j]));
                        };
                        break;
                    case google.maps.drawing.OverlayType.RECTANGLE:
                        bounds.extend(new google.maps.LatLng(_shapes[i].geomarker.northEast));
                        bounds.extend(new google.maps.LatLng(_shapes[i].geomarker.southWest));
                        break;
                    default:
                        throw 'Unknown shape type: ' + _shapes[i].type + '.';
                        break;
                };
            };

            var
                zoomChangeBoundsListener = google.maps.event.addListener(_map, 'bounds_changed', function() {
                    google.maps.event.removeListener(zoomChangeBoundsListener);
                    _map.setZoom(Math.min(minimumZoom, _map.getZoom()));
                });

            setTimeout(function() {
                google.maps.event.trigger(_map, 'resize');
                _map.setCenter(bounds.getCenter());
                _map.fitBounds(bounds);
                _map.setZoom(Math.min(minimumZoom, _map.getZoom()));
            }, 300);

            return _self;
        };

        /**
         * Get configuration parameter value.
         *
         * @param {string} key Parameter name
         * @return {*} Configuration value.
         */
        function getConfiguration(key) {
            return _settings[key];
        };

        /**
         * Set configuration parameter values.
         *
         * @param configuration
         * @return {GeomarkableMap}
         */
        function setConfiguration(configuration) {

            for (var property in configuration) {

                if (!_settings.hasOwnProperty(property)) {
                    throw 'Unknown configuration parameter "' + property + '" provided.';
                };

                _settings[property] = configuration[property];
            };

            return _self;
        }

        /**
         * Reset configuration to defaults.
         *
         * @return {GeomarkableMap}
         */
        function resetConfiguration() {

            for (var property in GeomarkableMap.defaults) {
                _settings[property] = GeomarkableMap.defaults[property];
            };

            return _self;
        };

        /**
         * Initialize
         */
        function init() {

            _self
                .resetConfiguration()
                .setConfiguration(settings);

            _map = new google.maps.Map(element, {
                center: {
                    lat: _settings.lat,
                    lng: _settings.lng
                },
                zoom:  _settings.zoom,
                mapTypeControl: _settings.mapTypeControl
            });

            _drawingManager = (_settings.editable) ? new google.maps.drawing.DrawingManager({
                drawingMode: null,
                drawingControl: false,
                markerOptions: {
                    draggable: true,
                    clickable: true
                },
                rectangleOptions: {
                    editable: true
                },
                circleOptions: {
                    editable: true
                },
                polygonOptions: {
                    editable: true
                },
                drawingControlOptions: {
                    drawingModes: [
                        google.maps.drawing.OverlayType.MARKER,
                        google.maps.drawing.OverlayType.CIRCLE,
                        google.maps.drawing.OverlayType.POLYGON,
                        google.maps.drawing.OverlayType.RECTANGLE
                    ]
                }
            }) : null;


            if (_settings.editable) {

                var
                    shapeToGeomarkerConverter = new ShapeToGeomarkerConverter();

                _drawingManager.setMap(_map);

                google.maps.event.addListener(_drawingManager, 'overlaycomplete', function(event) {

                    _self.enterDrawingMode(null);

                    var
                        shape = event.overlay;

                    shape.setMap(null);

                    var
                        shapeAdditionMethodName = 'add' + event.type.charAt(0).toUpperCase() +  event.type.toLowerCase().slice(1),
                        shapeToGeomarkerMethodName = event.type.toLowerCase() + 'ShapeToGeomarker';

                    if (!_self[shapeAdditionMethodName]) {
                        throw  'Unknown geomarker type: ' + event.type;
                    };

                    _self[shapeAdditionMethodName](shapeToGeomarkerConverter[shapeToGeomarkerMethodName](shape));
                });
            };
        };
    };

    function ShapeToGeomarkerConverter() {

        this.markerShapeToGeomarker = markerShapeToGeomarker;
        this.circleShapeToGeomarker = circleShapeToGeomarker;
        this.polygonShapeToGeomarker = polygonShapeToGeomarker;
        this.rectangleShapeToGeomarker = rectangleShapeToGeomarker;

        function markerShapeToGeomarker(shape) {

            return {
                lat: shape.position.lat(),
                lng: shape.position.lng(),
                type: google.maps.drawing.OverlayType.MARKER
            };
        };

        function circleShapeToGeomarker(shape) {

            return {
                center: {
                    lat: shape.center.lat(),
                    lng: shape.center.lng()
                },
                radius: shape.radius,
                type: google.maps.drawing.OverlayType.CIRCLE
            };
        };

        function polygonShapeToGeomarker(shape) {

            var
                points = [],
                polyShapes = shape.getPaths().getArray()[0].getArray();

            for (var i = 0; i < polyShapes.length; i++) {

                points.push({
                    lat: polyShapes[i].lat(),
                    lng: polyShapes[i].lng()
                });
            };

            return {
                points: points,
                type: google.maps.drawing.OverlayType.POLYGON
            };
        };

        function rectangleShapeToGeomarker(shape) {

            return {
                northEast: {
                    lat: shape.getBounds().getNorthEast().lat(),
                    lng: shape.getBounds().getNorthEast().lng()
                },
                southWest: {
                    lat: shape.getBounds().getSouthWest().lat(),
                    lng: shape.getBounds().getSouthWest().lng()
                },
                type: shape.type
            };
        };
    };


    // Make it public
    window.GeomarkableMap = GeomarkableMap;

})(window);
